#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <string.h>
#include <velib/types/ve_dbus_item.h>
#include <velib/types/ve_values.h>
#include <velib/canhw/canhw.h>
#include <velib/canhw/canhw_driver.h>
#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_items.h>
#include <velib/platform/console.h>
#include <velib/types/variant_print.h>
#include <velib/types/ve_dbus_item.h>
#include <velib/types/ve_item.h>
#include <velib/types/ve_item_def.h>
#include <velib/types/ve_values.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_item_utils.h>
#include <velib/vecan/products.h>

typedef enum {
	WRN_ONLY_SUB_RELAY_COMMAND,	/* byte 1 */
	BATTERY_HIGH_VOLTAGE,
	BATTERY_LOW_VOLTAGE,
	BATTERY_HIGH_TEMP,
	BATTERY_LOW_TEMP,
	BATTERY_HIGH_TEMPERATURE_CHARGE_ALARM,
	BATTERY_LOW_TEMPERATURE_CHARGE_ALARM,
	BATTERY_HIGH_CURRENT_DISCHARGE,
	BATTERY_HIGH_CURRENT_CHARGE,
	UNKNOWN_FLAG_BIT1_1,
	UNKNOWN_FLAG_BIT1_2,
	BMS_INTERNAL,
	CELL_IMBALANCE,
	ALARM_SUB_PACK2_ERROR,	// not a warning!
	ALARM_SUB_PACK1_ERROR,	// not a warning!
	UNKNOWN_FLAG_1_7
} LgAlarmFlags;

typedef enum {
	SONY_KEEPALIVE = 0x01
} Quirks;

typedef struct {
	VeItem soc;
	VeItem soh;
	VeItem voltage;
	VeItem current;
	VeItem power;
	VeItem temperature;
	VeItem maxChargeVoltaged;
	VeItem maxChargeCurrent;
	VeItem maxDischargeCurrent;
	VeItem batteryLowVoltage;
	VeItem ramWarnings;
	VeItem ramAlarms;
	VeItem productName;
	VeItem redetect;

	VeItem alarms[16];
	un32 bmsSeenCounter;
	struct VeItem *root;
	struct VeItem *productId;
	struct VeDbus *conn;
	un8 quirks;
	char manufacturerStr[9];
	char productNameStr[17];
} Device;

static Device theDevice;
static VeItem settingsProductId;
static un16 preferredProductId;
static char settingsPath[128];

static VeVariantUnitFmt unitVoltage1Dec = {1, "V"};
static VeVariantUnitFmt unitVoltage2Dec = {2, "V"};
static VeVariantUnitFmt unitPower = {0, "W"};
static VeVariantUnitFmt unitPercentage = {0, "%"};
static VeVariantUnitFmt unitAmps1Dec = {1, "A"};
static VeVariantUnitFmt unitCelsius1Dec = {1, "C"};

static void bmsPrecent(Device *dev);
static void registerBms(Device *dev, un16 productId, veBool forceProductId);
static void connectedChanged(Device *dev, veBool connected);
static void reconnect(Device *dev, un16 productId);
static void settingsProductIdChanged(struct VeItem *item);

/* Connect to the dbus and publish some values */
static void createDevice(struct VeItem *root, Device *dev)
{
	VeItem *battery, *alarms;
	VeVariant v;

	dev->root = root;

	dev->productId = veItemCreateProductId(root, VE_PROD_NOT_SET);
	veItemCreateBasic(root, "Connected", veVariantUn32(&v, 1));
	veItemCreateBasic(root, "Mgmt/Connection", veVariantStr(&v, "CAN-bus"));
	veItemCreateBasic(root, "DeviceInstance", veVariantUn16(&v, 0x200)); // ??
	veItemAddBasic(root, "ProductName", &dev->productName);
	veItemSet(&dev->productName, veVariantStr(&v, veProductGetName(VE_PROD_ID_GENERIC_BATTERY)));
	veItemAddBasic(root, "Redetect", &dev->redetect);
	veItemSet(&dev->redetect, veVariantSn8(&v, 0));

	veItemAddQuantity(root, "Soc", &dev->soc, &unitPercentage);
	veItemAddQuantity(root, "Soh", &dev->soh, &unitPercentage);

	/* Add values which should be accessible on the dbus */
	battery = veItemGetOrCreateUid(root, "Dc/0");
	veItemAddQuantity(battery, "Voltage", &dev->voltage, &unitVoltage2Dec);
	veItemAddQuantity(battery, "Current", &dev->current, &unitAmps1Dec);
	veItemAddQuantity(battery, "Power", &dev->power, &unitPower);
	veItemAddQuantity(battery, "Temperature", &dev->temperature, &unitCelsius1Dec);

	alarms = veItemGetOrCreateUid(root, "Alarms");
	veItemAddBasic(alarms, "LowVoltage", &dev->alarms[BATTERY_LOW_VOLTAGE]);
	veItemAddBasic(alarms, "HighVoltage", &dev->alarms[BATTERY_HIGH_VOLTAGE]);
	veItemAddBasic(alarms, "LowTemperature", &dev->alarms[BATTERY_LOW_TEMP]);
	veItemAddBasic(alarms, "HighTemperature", &dev->alarms[BATTERY_HIGH_TEMP]);
	veItemAddBasic(alarms, "HighDischargeCurrent", &dev->alarms[BATTERY_HIGH_CURRENT_DISCHARGE]);
	veItemAddBasic(alarms, "HighChargeCurrent", &dev->alarms[BATTERY_HIGH_CURRENT_CHARGE]);
	veItemAddBasic(alarms, "HighChargeTemperature", &dev->alarms[BATTERY_HIGH_TEMPERATURE_CHARGE_ALARM]);
	veItemAddBasic(alarms, "LowChargeTemperature", &dev->alarms[BATTERY_LOW_TEMPERATURE_CHARGE_ALARM]);
	veItemAddBasic(alarms, "CellImbalance", &dev->alarms[CELL_IMBALANCE]);
	veItemAddBasic(alarms, "InternalFailure", &dev->alarms[BMS_INTERNAL]);

	veItemAddQuantity(root, "Info/MaxChargeVoltage", &dev->maxChargeVoltaged, &unitVoltage1Dec);
	veItemAddQuantity(root, "Info/MaxChargeCurrent", &dev->maxChargeCurrent, &unitAmps1Dec);
	veItemAddQuantity(root, "Info/MaxDischargeCurrent", &dev->maxDischargeCurrent, &unitAmps1Dec);
	veItemAddQuantity(root, "Info/BatteryLowVoltage", &dev->batteryLowVoltage, &unitVoltage1Dec);

}

/* Called after CAN is already setup */
void taskInit(void)
{
	/* normally this is done after a device is found on the CAN bus */
	VeItem *producer;
	VeItem *consumer;
	const char *settingsService = "com.victronenergy.settings";

	producer = veItemGetOrCreateUid(veValueTree(), "com.victonenergy.producer");
	createDevice(producer, &theDevice);

	/* Connect to the dbus */
	struct VeDbus *conn = veDbusConnectString(veDbusGetDefaultConnectString());
	if (!conn) {
		printf("dbus connection failed\n");
		pltExit(12);
	}

	/* Listen to D-Bus.. */
	veDbusSetListeningDbus(conn);

	/* Connect to settings service */
	consumer = veItemGetOrCreateUid(veValueTree(), "com.victonenergy.consumer");
	if (!veDbusAddRemoteService(settingsService, consumer, veTrue)) {
		printf("veDbusAddRemoteService failed\n");
		pltExit(13);
	}

	{
		VeVariant defaultValue;
		VeVariant minValue;
		VeVariant maxValue;

		char buffer[32];
		char *c0;
		char *c1;
		veBool cap = veTrue;
		VeCanGateway *canGw = veCanGwActive();
		veCanGwId(canGw, buffer, sizeof(buffer));
		c0 = buffer;
		c1 = c0;
		for (; *c0 !=0 ; ++c0) {
			if (isalnum(*c0)) {
				*c1 = cap ? toupper(*c0) : *c0;
				++c1;
				cap = veFalse;
			} else {
				cap = veTrue;
			}
		}
		*c1 = 0;
		snprintf(settingsPath, sizeof(settingsPath), "Settings/CanBms/%s/ProductId", buffer);

		/* Create an item pointing to our new setting. The zero in the path indicates we are
		 * using the can0 interface. */
		veItemAddBasic(consumer, settingsPath, &settingsProductId);

		/* Set the properties of the new settings */
		veVariantSn32(&defaultValue, 0);
		veVariantSn32(&minValue, 0);
		veVariantSn32(&maxValue, 0);

		/* Set request to the localsettings service */
		if (!veDBusAddLocalSetting(&settingsProductId, &defaultValue, &minValue, &maxValue, veFalse)) {
			printf("veDBusAddLocalSetting failed\n");
			pltExit(14);
		}

		veItemSetChanged(&settingsProductId, settingsProductIdChanged);
	}
}

// helper, see 35A
static void parseAlarm(VeItem *item, un32 *combinedFlags)
{
	VeVariant v;
	if (item != NULL)
		veItemOwnerSet(item, veVariantUn8(&v, *combinedFlags & 0x3));
	*combinedFlags >>= 2;
}

/**
 * Reads a null terminated string from the CAN-bus stream.
 * @param str the buffer where the string will be stored. str will always be null terminated, so
 * make sure to allocated an extra character in case the stream itself does not contain a zero.
 * @return The length of the string (null terminator excluded).
 */
static int veInStr(char *str)
{
	char *s = str;
	for (;;) {
		char c = (char)veInUn8();
		if (isprint(c)) {
			*s = c;
			++s;
		}
		if (c == 0)
			break;
	}
	*s = 0;
	return s - str;
}

void settingsProductIdChanged(struct VeItem *item)
{
	preferredProductId = (un16)item->variant.value.UN32;
	printf("Product ID from settings: %d\n", (int)preferredProductId);
}

void taskUpdate(void)
{
	VeRawCanMsg msg;
	Device *dev = &theDevice;

	while (veCanRead(&msg))
	{
		VeStream canIn;

		/* ignore all 29 bit / extended messages */
		if (msg.flags & VE_CAN_EXT)
			continue;

		/* note: io has a bunch of helpers to parse a byte array */
		veInStreamInit(&canIn, msg.mdata, msg.length);
		switch (msg.canId) {
		case 0x351:
			/*
			 * Battery Charge Voltage
			 * (data type : 16bit unsigned int, byte order : little endian, scale factor : 0.1, unit : V)"
			 *
			 * DC Charge Current Limitation
			 * (data type : 16bit signed int, 2's complement, byte order : little endian, scale factor : 0.1, unit : A)"
			 *
			 * DC Discharge Current Limitation
			 * (data type : 16bit signed int, 2's complement, byte order : little endian, scale factor : 0.1, unit : A)"
			 *
			 * Battery discharge voltage U16 0.1V Voltage discharge limit
			 */
			veInItemLe(&dev->maxChargeVoltaged, VE_SN16, 1);
			veInItemLe(&dev->maxChargeCurrent, VE_SN16, 1);
			veInItemLe(&dev->maxDischargeCurrent, VE_SN16, 1);
			veInItemLe(&dev->batteryLowVoltage, VE_UN16, 1);
			bmsPrecent(dev);
			break;
		case 0x355:
			/*
			 * SOC Value
			 * (data type : 16bit unsigned int, byte order : little endian, scale factor : 1, unit : %)"
			 *
			 * SOH Value
			 * (data type : 16bit unsigned int, byte order : little endian, scale factor : 1, unit : %)"
			 */
			veInItemLe(&dev->soc, VE_UN16, 0);
			veInItemLe(&dev->soh, VE_UN16, 0);
			bmsPrecent(dev);
			break;
		case 0x356:
			/*
			 * Battery Voltage
			 * (data type : 16bit signed int, 2's complement, byte order : little endian, scale factor : 0.01, unit : V)"
			 *
			 * Battery Current
			 * (data type : 16bit signed int, 2's complement, byte order : little endian, scale factor : 0.1, unit : A)"
			 *
			 * Battery Temperature
			 * (data type : 16bit signed int, 2's complement, byte order : little endian, scale factor : 0.1, unit : degC)"
			 */
			veInItemLe(&dev->voltage, VE_SN16, 2);
			veInItemLe(&dev->current, VE_SN16, 1);
			veItemCalculatePower(&dev->voltage, &dev->current, &dev->power);
			veInItemLe(&dev->temperature, VE_SN16, 1);
			bmsPrecent(dev);
			break;
		case 0x359:
			/* LG resu, Sony & Pylontech send these */
			{
				un16 warnings;
				un16 alarms;
				int n;
				VeVariant v;
				un16 mask = B1b00011001 | b10011110; /* both in warning AND alarms */
				un16 productId;

				alarms = veInUn16Le();
				warnings = veInUn16Le();
				logI("lg", "warnings %04X, alarms %04X", warnings, alarms);

				productId = VE_PROD_ID_LG_RESU_BATTERY;
				if (msg.length >= 7) {
					// Pylontech batteries send an ID in bytes 5 & 6
					veInUn8();
					if (veInUn16Le() == 0x4E50)
						productId = VE_PROD_ID_PYLONTECH_BATTERY;
				}
				registerBms(dev, productId, productId != VE_PROD_ID_LG_RESU_BATTERY);

				/* remove the special flags */
				warnings &= mask;
				alarms &= mask;

				for (n = 0; n < 16; n++) {
					un8 value = (alarms & 1 ? 2 : (warnings & 1 ? 1 : 0));
					veItemOwnerSet(&dev->alarms[n], veVariantUn8(&v, value));
					warnings >>= 1;
					alarms >>= 1;
				}
			}
			bmsPrecent(dev);
			break;
		case 0x35A:
			{
				/*
				 * Each alarms and warning has a seperate bit to set and clear the alarm
				 * 1 -> raised and 2 is cleared. 0 and 3 are ignored. Only the raised values
				 * are forwarded.
				 */
				un32 alarms = veInUn32Le() & 0x55555555UL;
				un32 warnings = veInUn32Le() & 0x55555555UL;
				un32 combined = (alarms << 1) | warnings;

				registerBms(dev, VE_PROD_ID_GENERIC_BATTERY, veFalse);

				parseAlarm(NULL, &combined); /* external alarm 1 */
				parseAlarm(&dev->alarms[BATTERY_HIGH_VOLTAGE], &combined);
				parseAlarm(&dev->alarms[BATTERY_LOW_VOLTAGE], &combined);
				parseAlarm(&dev->alarms[BATTERY_HIGH_TEMP], &combined);
				parseAlarm(&dev->alarms[BATTERY_LOW_TEMP], &combined);
				parseAlarm(&dev->alarms[BATTERY_HIGH_TEMPERATURE_CHARGE_ALARM], &combined); /* too hot to charge */
				parseAlarm(&dev->alarms[BATTERY_LOW_TEMPERATURE_CHARGE_ALARM], &combined); /* too cold to charge */
				parseAlarm(&dev->alarms[BATTERY_HIGH_CURRENT_DISCHARGE], &combined);
				parseAlarm(&dev->alarms[BATTERY_HIGH_CURRENT_CHARGE], &combined);
				parseAlarm(NULL, &combined); /* Technical problems with the contacter */
				parseAlarm(NULL, &combined); /* Short circuit within the battery */
				parseAlarm(&dev->alarms[BMS_INTERNAL], &combined);
				parseAlarm(&dev->alarms[CELL_IMBALANCE], &combined);
				/* reserved */
				/* reserved */
				/* generator */
			}
			bmsPrecent(dev);
			break;
		case 0x35E:
			if (*dev->manufacturerStr != 0)
				break;
			/* Sony will send this message roughly every 5.5 seconds */
			veInStr(dev->manufacturerStr);
			printf("Manufacturer: '%s'\n", dev->manufacturerStr);
			// printf("Manufacturer: '%s'\n", dev->manufacturerStr);
			// The manufacturer provided by the puredrive battery (PDSOLAR) may contains an extra
			// space at the end, so we use strstr instead of strcmp.
			if (strcmp(dev->manufacturerStr, "Sony") == 0) {
				reconnect(dev, VE_PROD_ID_SONY_BATTERY);
			} else if (strstr(dev->manufacturerStr, "PDSOLAR") != NULL) {
				reconnect(dev, VE_PROD_ID_PURE_DRIVE_BATTERY);
			} else if (strcmp(dev->manufacturerStr, "BYD") == 0) {
				reconnect(dev, VE_PROD_ID_BYD_BATTERY);
			}
			break;
		case 0x370:
			if (*dev->productNameStr == 0)
				veInStr(dev->productNameStr);
			break;
		case 0x371:
			if (*(dev->productNameStr + 8) == 0)
				veInStr(dev->productNameStr + 8);
			if (*dev->productNameStr != 0) {
				VeVariant v;
				veItemSet(&dev->productName, veVariantStr(&v, dev->productNameStr));
				break;
			}
		}
	}
}

static veBool sendPresence(void)
{
	VeRawCanMsg msg;

	if (theDevice.quirks & SONY_KEEPALIVE) {
		msg.length = 6;
		msg.canId = 0x306;
	} else {
		msg.length = 8;
		msg.canId = 0x305;
	}
	memset(&msg.mdata[0], 0, msg.length);
	msg.flags = 0;

	return veCanSend(&msg);
}

static void connectedChanged(Device *dev, veBool connected)
{
	if (connected) {
		char serviceName[256];
		char buffer[32];
		char *c;

		dev->conn = veDbusConnectString(veDbusGetDefaultConnectString());

		if (dev->conn == NULL) {
			printf("dbus_service: no dbus connection\n");
			pltExit(10);
		}

		strcpy(serviceName, "com.victronenergy.battery.");
		VeCanGateway *canGw = veCanGwActive();
		veCanGwId(canGw, buffer, sizeof(buffer));
		for (c = buffer; *c != 0; ++c) {
			if (!isalnum(*c))
				*c= '_';
		}
		strcat(serviceName, buffer);

		printf("dbus_service: registering name %s\n", serviceName);
		veDbusItemInit(dev->conn, dev->root);

		/* make the values available on the dbus and get a service name */
		if (!veDbusChangeName(dev->conn, serviceName)) {
			printf("dbus_service: registering name failed\n");
			pltExit(11);
		}
	} else {
		printf("disconnecting from dbus\n");
		veDbusDisconnect(dev->conn);
		dev->conn = NULL;
		theDevice.quirks = 0;
		memset(theDevice.manufacturerStr, 0, sizeof(theDevice.manufacturerStr));
		memset(theDevice.productNameStr, 0, sizeof(theDevice.productNameStr));
	}
}

static void bmsPrecent(Device *dev)
{
	dev->bmsSeenCounter = 3;
}

/*
 * note: hacky way of determining the battery type, the BMZ and LG resu send different
 * alarm messages, see 359 / 35A.
 */
static void registerBms(Device *dev, un16 productId, veBool forceProductId)
{
	VeVariant v;
	const char *productName = NULL;
	if (dev->conn != NULL)
		return;
	if (!veItemIsValid(&settingsProductId)) {
		printf("ProductId setting not valid\n");
		return;
	}
	if (preferredProductId == 0 || forceProductId) {
		preferredProductId = productId;
	} else {
		productId = preferredProductId;
	}
	veItemSet(&settingsProductId, veVariantUn32(&v, productId));
	if (productId == VE_PROD_ID_SONY_BATTERY)
		dev->quirks |= SONY_KEEPALIVE;
	veItemSet(dev->productId, veVariantUn16(&v, productId));
	if (!*dev->productNameStr) {
		productName = veProductGetName(productId);
		veItemSet(&dev->productName, veVariantStr(&v, productName));
	}
	connectedChanged(dev, veTrue);
}

static void reconnect(Device *dev, un16 productId)
{
	veAssert(veItemIsValid(&settingsProductId));
	if (dev->conn == NULL) {
		preferredProductId = productId;
		return;
	}
	if (preferredProductId == productId)
		return;
	connectedChanged(dev, veFalse);
	registerBms(dev, productId, veTrue);
}

/* 50ms time progress */
void taskTick(void)
{
	static un8 sec;

	if (theDevice.redetect.variant.value.UN8 == 1) {
		VeVariant v;
		veItemSet(&theDevice.redetect, veVariantSn8(&v, 0));
		preferredProductId = 0;
		connectedChanged(&theDevice, veFalse);
		return;
	}

	if (sec == 20) {
		if (sendPresence())
			sec = 0;
		if (theDevice.bmsSeenCounter && --theDevice.bmsSeenCounter == 0)
			connectedChanged(&theDevice, veFalse);
	} else {
		sec++;
	}
}

char *pltProgramVersion(void)
{
	return "v0.10";
}
